package org.ksv.plugin;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;

import org.eclipse.jface.action.IAction;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.text.templates.Template;
import org.eclipse.jface.text.templates.persistence.TemplatePersistenceData;
import org.eclipse.jface.text.templates.persistence.TemplateStore;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.IWorkbenchWindowActionDelegate;

import com.openedge.pdt.text.OEProject;
import com.openedge.pdt.text.template.OETemplateSettings;

/**
 * Converts p4gl aliases from OE Studio into Eclipse's templates
 * @author ksv
 *
 */
public class ImportAliases implements IWorkbenchWindowActionDelegate, IKsvPlugin{

	private IWorkbenchWindow window;
	
	public void dispose() {
		// TODO Auto-generated method stub		
	}

	public void init(IWorkbenchWindow window) {
		this.window = window;
		
	}

	public void run(IAction action) {
		KsvUtils.runPlugin(this);		
	}
	
	public void selectionChanged(IAction action, ISelection selection) {		
	}
	
	/**
	 * Converts p4gl aliases from OE Studio into Eclipse's templates
	 */
	public void runAction(OEProject project) {	
		
		// Open the template store
		_templateStore = OETemplateSettings.getTemplateStore();
		
		// File with OES aliases 
		_p4glalsFile = "p4gl.als";	
		
		
		try {
			// Ask for an alias file
			_p4glalsFile = KsvUtils.openFileDialog(window,"Choose 4gl alias file",new String[] { _p4glalsFile, "*.als" }, _p4glalsFile );

			if ( _p4glalsFile == null ) return;
			
			// Parse alias file and add aliases into Eclipse's templates
			ParseAliasFile();
			
			// Make changes persistent
			_templateStore.save();
			
		}
		catch( Exception e ) {		
			MessageDialog.openError( window.getShell(), "KsvPlugin", " Aliases haven't been imported. \n\n" + e.getMessage() );
		}
		
		MessageDialog.openInformation( window.getShell(), "KsvPlugin", " Aliases have been imported succesfully." );
		
	}
	
	/**
	 * Parses the alias file and add the aliases into XML
	 * @throws IOException 
	 */
	private void ParseAliasFile() throws IOException {
		BufferedReader breader = null;
		try {
			
			breader = new BufferedReader( new FileReader( _p4glalsFile ) );
			String alias = "";
			int trimPos = 0;
			
			try {
				// Read the alias file string by string 
				while( breader.ready() ) {
					String line = breader.readLine();
					
					// Ignore empty lines
					if ( line.equals( "" ) ) continue;
					
					// if a string doesn't begin with space 
					// it's a beginning of a new alias 
					if ( !line.startsWith(" ") && !alias.equals("") ) {
						ParseAlias( alias );
						alias = "";
					}

					// the beginning of an alias
					// a body of an alias has indentation at the beginning 
					// which should be cut. how many spaces should be cut
					// depends on either the length of the alias name or
					// whether the alias has parameters 
					if ( alias.equals( "" ) ) {					
						// is it parametrized? 
						if ( line.indexOf( "(" ) >= 0 ) {
							trimPos = 1;
						}
						else {
							trimPos = line.indexOf( " " ) + 1;
						}				
					}
					else {
						line = line.substring( trimPos );
					}


					alias += ( alias == "" ? "" : "\n" ) + line;
				}	
				
				if ( !alias.equals("") ) {
					ParseAlias( alias );
				}
			}
			finally {
				breader.close();				
			}
		}
		catch ( IOException e ) {			
		}			
	}

	/**
	 * Parses an alias string and adds it into XML file if it's needed
	 * @param alias
	 */
	private void ParseAlias(String alias) {
		int spacePos = alias.indexOf( " " );
		String name = alias.substring( 0, spacePos ).toUpperCase();
		
		int obracketPos = name.indexOf( "(" );
		
		// The open bracket says that the alias has parameters
		// which should be ignored
		if ( obracketPos >= 0 ) {
			name = name.substring( 0, obracketPos );
		}
		
		// There's such an alias in XML
		if ( _templateStore.findTemplate( name ) != null ) return;
		
		// Cut parameters in the alias's body:
		// find the close bracket and then the first space
		if ( obracketPos >= 0 ) {
			spacePos = alias.indexOf( ")" );
			spacePos = alias.indexOf( "\n", spacePos );
		}
		
		String body = alias.substring( spacePos + 1 );
		
		// alias can contain spaces at the beginning which don't matter
		// so they have to be removed for each line
		int falseIndent = 0;
		// count spaces at the beginning
		while( falseIndent < body.length() && body.charAt( falseIndent++ ) == ' '  );
		// remove false spaces from every line		
		body = body.substring( falseIndent - 1 );
		body = body.replaceAll( "\n\\x20{" + (falseIndent - 1) + "}", "\n");
		
		// Clean the body from alias things such as parameters
		// and internal variables
		body = body.replace( "%\\c", "${cursor}" );
		body = body.replace( "%\\S", "" );
		body = body.replace( "%\\i", "" );
		body = body.replaceAll("\\%\\(\\w+\\)", "" );
		
		// Add a converted macro into template store
		Template template = new Template( name, 
				"Imported by KsvPlugin", 
				"com.openedge.pdt.text.editor.templates",
				body, 
				true);
		_templateStore.add( new TemplatePersistenceData( template, true ) );
	}	
	
	TemplateStore _templateStore;
	private String _p4glalsFile = "p4gl.als";
}
