package org.ksv.plugin;

import java.io.File;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.core.filesystem.EFS;
import org.eclipse.core.filesystem.IFileStore;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.IDocumentExtension3;
import org.eclipse.jface.text.IDocumentPartitioner;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.rules.FastPartitioner;
import org.eclipse.swt.widgets.DirectoryDialog;
import org.eclipse.swt.widgets.FileDialog;
import org.eclipse.ui.IEditorDescriptor;
import org.eclipse.ui.IEditorInput;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IEditorReference;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.IWorkbenchWindow;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.PlatformUI;
import org.eclipse.ui.ide.FileStoreEditorInput;
import org.eclipse.ui.ide.IDE;
import org.eclipse.ui.part.FileEditorInput;
import org.eclipse.ui.texteditor.ITextEditor;

import com.openedge.pdt.core.OECommonServicesPlugin;
import com.openedge.pdt.text.OEPlugin;
import com.openedge.pdt.text.OEProject;
import com.openedge.pdt.text.OEutil;
import com.openedge.pdt.text.editor.OETextEditor;
import com.openedge.pdt.text.editor.rules.ABLPartitionScanner;
import com.openedge.pdt.text.project.ProjectPropath;
import com.openedge.pdt.text.project.PropathEntry;

/**
 * Utility class with necessary plugin functions
 * @author ksv
 *
 */
public class KsvUtils {
	/**
	 * Gets current OpenEdge project using either the selection or
	 * the file opened in the active editor
	 * @param selection 
	 * @param window 
	 * @return
	 */
	@SuppressWarnings("deprecation")
	public static OEProject getActiveOEProject()
	{
		IProject project = getActiveProject();		
		
		if(project == null)
			return null;
		else
			return OEProject.getOEProject(project);
	}
	
	/**
	 * Gets current project using either the selection or
	 * the file opened in the active editor
	 * @return
	 */
	public static IProject getActiveProject() {
//		IResource resource = null;
//
//		resource = OECommonServicesPlugin.getSelectedResource();
//
//		if (resource == null)
//			resource = OECommonServicesPlugin.getEditorResource();
//
//
//		if(resource == null)
//			return null;
//		else
//			return resource.getProject();
		return OEutil.getProjectSelection();		
	}
	
	
	/**
	 * Opens the common file dialog to select a file
	 * @param window       	- current window
	 * @param title			- title of the dialog
	 * @param fileFilters	- file filters
	 * @param filename		- initial filename
	 * @return filename or null if a user clicked cancel
	 */
	public static String openFileDialog(IWorkbenchWindow window,String title,String[] fileFilters, String filename) {
		FileDialog dialog = new FileDialog(window.getShell());
		dialog.setFilterExtensions(fileFilters);
		dialog.setText(title);
		dialog.setFileName(filename);
		return dialog.open();
	}	
	
	/**
	 * Opens the common folder dialog to select a folder
	 * @param window - current window
	 * @param title  - title of the dialog
	 * @return a folder or null if a user clicked cancel
	 */
	public static String openFolderDialog ( IWorkbenchWindow window, String title ) {
		DirectoryDialog dialog = new DirectoryDialog( window.getShell() );
		dialog.setText( title );
		return dialog.open( );		
	}
	
	
	/**
	 * Detects if the file is an ABL file
	 * @param file
	 * @return
	 */
	public static boolean isProgressFile( File file ) {
		IEditorDescriptor desc = PlatformUI.getWorkbench().getEditorRegistry().getDefaultEditor( file.getName() );
		if ( desc == null ) 
			return false;
		else {
			tmpEditorId = desc.getId();
			return tmpEditorId.startsWith( "com.openedge.pdt.text.editor.OEEditor" );
		}
	}
	
	private static boolean isCopiedFile( File file ) {		
		try {
			IFileStore tempStore = getTempFileStore( file );
			File tempFile = tempStore.toLocalFile( EFS.NONE, null );
			if ( tempFile != null )
				return tempFile.getAbsolutePath().equals( file.getAbsolutePath() );
			else
				return false;
		}
		catch ( CoreException e ) {
			return false;
		}		
	}
	
	/**
	 * Opens an external or internal file in the editor
	 * @param fileToOpen - file to open	 * 
	 * @return true - if file's been opened successfully
	 */
	public static boolean openFileInEditor( File fileToOpen ) {					
		
		if ( isCopiedFile( fileToOpen ) ) {
			fileToOpen = lookForFileInPropath( lastProject, fileToOpen.getName() );
			if ( !fileToOpen.exists() ) return false;
		}
		
		// Check whether the file we want to open is in the project.
		// If it's not, open it in read-only mode
		boolean readonly = lastProject.file2resource( fileToOpen.getAbsolutePath() ) == null;

		
	    try {	
	    	
	    	// There's no simple way to make the editor readonly, so we make a copy of the file readonly
	    	if ( readonly ) {
	    		fileToOpen = setFileReadonly( fileToOpen );
	    		if ( fileToOpen == null ) return false;
	    	}
	    	
	    	// Get id of editor for current file
    		boolean highlightNeeded = false;
	    	
    		String editorId = "";
    		// if it's not an ABL file, use ABL editor for includes
    		if ( !isProgressFile( fileToOpen ) ) {
	    		editorId = OETextEditor.I_EDITOR_ID;
	    		highlightNeeded = true;
    		}
    		else {
    			editorId = tmpEditorId;
    		}    		
	    	
    		IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
    		
	    	IEditorPart editorPart = IDE.openEditor( page, fileToOpen.toURI( ), editorId, true );
	    	
	    	if ( highlightNeeded ) restoreAblHighlighting( editorPart );

	    } catch ( PartInitException e ) {
	    	return false;	        
	    }		
	    return true;
	}
	
	/**
	 * Returns a temp file for the given file 
	 * @param file
	 * @return
	 * @throws CoreException
	 */
	private static IFileStore getTempFileStore( File file ) throws CoreException {		
		// Form a new file name in the temp directory
		IPath newPath = Activator.getDefault().getStateLocation().addTrailingSeparator().append( file.getName() );
		return EFS.getStore( newPath.toFile().toURI() );		
	}

	/**
	 * Copies file into the temp location, make it readonly, and 
	 * returns the copied file
	 * @param file
	 * @return
	 */
	private static File setFileReadonly( File file ) {
		try {
			IFileStore srcFile = EFS.getStore( file.toURI() );
			// Form a new file name in the temp directory			
			IFileStore tmpStore = getTempFileStore( file );
			File tmpFile = tmpStore.toLocalFile( EFS.NONE, null );
			
			// Check whether the file is already in the temp dir
			// If it is, delete it
			if ( tmpFile.exists() ) {
				tmpFile.delete();		    			
			}
			
			// Copy the file 
			srcFile.copy( tmpStore, EFS.OVERWRITE, null );
			
			// Add the original file into history
			CopyHistory.setCopyHistory( tmpFile, file );
			
			file = tmpFile;
			// Make it readonly
			file.setReadOnly();	    			
		}
		catch ( CoreException e ) {
			return null;
		}
		return file;
	}
	
	/**
	 * Returns cursor position in the current editor
	 * @return
	 */
	public static int getCurrentEditorOffset() {
		IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();	
		ITextEditor editor = ( ITextEditor ) page.getActiveEditor();
		if ( editor != null ) {
			ITextSelection selection = (ITextSelection) editor.getSelectionProvider().getSelection();
			if ( selection != null ) 
				return selection.getOffset();
		}
		return -1;		
	}
	
	/**
	 * Sets cursor in the given position
	 * @param offset - position of cursor
	 */
	public static void setCurrentEditorOffset( int offset ) {
		IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();	
		ITextEditor editor = ( ITextEditor ) page.getActiveEditor();
		if ( editor != null ) {			
			editor.setHighlightRange( offset, 0, true);
		}			
	}	

	/**
	 * Closes file opened in the active editor
	 * @return true if file has been closed OK
	 */
	public static boolean closeFileInCurrentEditor() {
		IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
		return page.closeEditor(page.getActiveEditor(), false);		
	}
	
	/**
	 * Reopens a file
	 * @param file
	 */
	public static void reopenFile(File file) {
		// Save the current cursor position
		int offset = getCurrentEditorOffset();
		if ( closeFileInCurrentEditor() ) {
			openFileInEditor( file );
			// Restore the cursor position
			if ( offset >= 0 ) {
				setCurrentEditorOffset( offset );							
			}
		}
	}
	
	/**
	 * Opens an external file in the editor as a link
	 * @param fileToOpen - file to open
	 * @param readonly - open file in readonly mode
	 * @return true - if file's been opened successfully
	 */
	public static boolean openFileInEditorAsLink(File fileToOpen, boolean readonly) {
		boolean result = true;

		try {
			IPath location = new Path(fileToOpen.getAbsolutePath());
			IFile file = lastProject.getProject().getFile(location.lastSegment());
			file.createLink(location, IResource.FILE, null);
			// if (readonly) file.setReadOnly(true);
			IWorkbenchPage page = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage();
			if (page != null) IDE.openEditor(page, file);
		}
		catch (Exception e) {
			result = false;	
		}		
		return result;		
	}
	
	/**
	 * Looks for a file using PROPATH
	 * @param project - OpenEdge project
	 * @param filename - a file to find
	 * @return file
	 */
	public static File lookForFileInPropath(OEProject project, String filename) {
		IPath file = project.getPropath().search(filename, false);
		if (file != null) 
			return file.toFile();
		else
			return new File(filename);
	}
	
	/**
	 * Gets file opened in the editor even if the file doesn't belong to the workspace
	 * @param window
	 * @return File
	 */
	public static File getEditorFile(IWorkbenchWindow window) {
		IPath path = OECommonServicesPlugin.getEditorPath();
		
		if (path == null) {
			IEditorPart editorPart = window.getActivePage().getActiveEditor();
			if(editorPart == null) return null;
			FileStoreEditorInput fileStore = (FileStoreEditorInput) editorPart.getEditorInput();
			return new File(fileStore.getURI());
		}
		return path.toFile();
	}
	

	/**
	 * Runs plugin action
	 * @param plugin
	 */
	public static void runPlugin(IKsvPlugin plugin) {
		// Get the current OE project
		OEProject project = getActiveOEProject();
		
		// If the current project isn't openedge's try the last openedge's one
		if (project == null) 
			project = KsvUtils.lastProject;

		if (project != null) {
			
			// Remember the openedge project
			KsvUtils.lastProject = project;
			
			plugin.runAction(project);
		}
	}
	
	/**
	 * Opens a file using OpenEdge PROPATH
	 * @param filename 		- file to open
	 * @param project  		- OpenEdge project whose PROPATH should be used
	 * @param showMessage	- show any message
	 * @param window		- current window
	 * @return
	 */
	public static boolean openFileInPropath (String filename, OEProject project, boolean showMessage,IWorkbenchWindow window) {
		boolean fileOpened = false;
		
		if ( filename.equals( "" ) ) return false;
		
		File fileToOpen = new File(filename);
		// If user has typed an unreal name, try to find it using PROPATH
		// at first we use a pathname 
		if ( !fileToOpen.exists() ) {
			fileToOpen = lookForFileInPropath ( project, fileToOpen.getPath() );
			// pathname didn't work, use filename
			if ( !fileToOpen.exists() ) {
				fileToOpen = lookForFileInPropath ( project, fileToOpen.getName() );
			}
		}
		
		// If there's anything what can be opened, open it
		if ( fileToOpen.exists() )				
			fileOpened = openFileInEditor( fileToOpen );	
		else
		{
			if (showMessage) {
				MessageDialog.openInformation(
						window.getShell(), 
						"KsvPlugin", 
						"File [" + fileToOpen.getName() + "] cannot be found.");
			}
		}
		return fileOpened;
	}
	
	/**
	 * Add dirs into project's propath
	 * @param project 	- OE project
	 * @param pathList	- list of folder to add into PROPATH
	 * @param save		- make PROAPTH persistent
	 */
	public static void addDirsToPropath ( OEProject project, String pathList, boolean save ) {
		
		ProjectPropath propath = project.getPropathHandler();
		String[] paths = pathList.split(",");		
				
		List<PropathEntry> propathEntries = new ArrayList<PropathEntry>();
		Collections.addAll( propathEntries, propath.getPropathEntries() );
		
		for( String path : paths ) {
			// Check before adding if the path is already in the list 
			boolean found = false;
			for (PropathEntry pEntry : propathEntries ) {
				if ( pEntry.getPath().equalsIgnoreCase( path ) ) {
					found = true;
					break;
				}				
			}
			// If path isn't in PROPATH
			if ( !found ) {
				PropathEntry entry = new PropathEntry( "dir" );		
				entry.setPath( path );		
				propathEntries.add( entry );
			}
		}
		// Change PROPATH and save it for the next runs
		propath.setPropathEntries( propathEntries.toArray( new PropathEntry[propathEntries.size()] ) );
		if ( save ) propath.savePropathFile();
	}
	
	/**
	 * Set new PROPATH for project
	 * @param project	- a OE project
	 * @param pathList	- list of folder for PROPATH
	 * @param save		- make PROAPTH persistent
	 */
	public static void setPropath ( OEProject project, String pathList, boolean save ) {
		// Reset the current PROPATH		
		ProjectPropath propath = project.getPropathHandler();
		propath.setPropathEntries( null );
		// Set a new PROPATH
		addDirsToPropath( project, pathList, save );				
	}

	/**
	 * Restores ABL syntax highlighting for the editor
	 * @param editorPart - editor
	 */
	public static void restoreAblHighlighting( IEditorPart editorPart ) {		
		if ( editorPart instanceof OETextEditor ) {		
			OETextEditor editor = ( OETextEditor ) editorPart;
			IDocument document = editor.getDocument();
			IDocumentPartitioner partitioner = new FastPartitioner( OEPlugin.getDefault().getABLPartitionScanner(), 
					ABLPartitionScanner.ABL_PARTITION_TYPES);
			partitioner.connect( document );
			( ( IDocumentExtension3 ) document ).setDocumentPartitioner( OEPlugin.ABL_PARTITIONING, partitioner) ;			
		}		
	}
	
	/**
	 * Restores ABL syntax highlighting for all ABL editors 
	 * containing non ABL files at startup
	 */
	public static void restoreAblHighligtingAtStartup() {
		// Get all editors
		IEditorReference[] references = PlatformUI.getWorkbench().getActiveWorkbenchWindow().getActivePage().getEditorReferences();
		for ( IEditorReference reference : references ) {
			IEditorPart editor = reference.getEditor( true );
			// If it's the ABL editor 
			if ( editor instanceof OETextEditor ) {				
				IEditorInput input = editor.getEditorInput();
				File file = null;
				// Get the file opened in the editor
				if ( input instanceof FileEditorInput ) {
					FileEditorInput fileInput = ( FileEditorInput ) input;
					file = fileInput.getPath().toFile();
				}
				if ( input instanceof FileStoreEditorInput ) {
					FileStoreEditorInput fileStore = ( FileStoreEditorInput ) input;
					file = new File( fileStore.getURI() );
				}
				// Is the file an ABL file? 
				// If it's not, restore its highlighting
				if ( !isProgressFile( file ) ) {
					restoreAblHighlighting( editor );				
				}
			}
		}
	}
	
	public static OEProject getLastProject() {
		return lastProject;
	}
	
	static String tmpEditorId = "";
	static OEProject lastProject;	
	
}
